from jinja2 import Environment, FunctionLoader, pass_context

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.builders.utils import ensure_directory
from oarepo_model_builder.templates import templates


class InvenioScriptBootstrapBuilder(OutputBuilder):
    TYPE = "invenio_script_bootstrap"

    def finish(self):
        context = {"settings": self.schema.settings}

        env = Environment(
            loader=FunctionLoader(lambda tn: templates.get_template(tn, context["settings"])),
            autoescape=False,
        )

        ensure_directory(self.builder, "scripts")
        output = self.builder.get_output("diff", "scripts/bootstrap.sh")
        output.write(env.get_template("script-bootstrap").render(context))
        output.make_executable()



from collections import defaultdict

from oarepo_model_builder.builders import process

from ..utils.deepmerge import deepmerge
from ..utils.jinja import base_name, package_name, with_defined_prefix
from .invenio_base import InvenioBaseClassPythonBuilder

OAREPO_MARSHMALLOW_PROPERTY = "oarepo:marshmallow"


class MarshmallowNode:
    def __init__(self, schema_class_name, schema_class_bases, schema):
        self.schema_class_name = schema_class_name
        self.schema_class_bases = schema_class_bases
        self.schema = schema
        self.fields = {}
        self.prepared_schema = None

    def add(self, field_name, field_type):
        self.fields[field_name] = field_type

    def pop_prepared_schema(self):
        ret = self.prepared_schema
        self.prepared_schema = None
        return ret


class InvenioRecordSchemaBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_schema"
    class_config = "record-schema-class"
    template = "record-schema"

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.marshmallow_stack = []
        self.imports = defaultdict(set)  # import => aliases
        self.imports["marshmallow"].add("ma")
        self.imports["marshmallow.fields"].add("ma_fields")
        self.imports["marshmallow.validate"].add("ma_valid")
        self.imported_classes = {}
        self.generated_classes = set()

    def finish(self):
        super().finish(imports=self.imports, imported_classes=self.imported_classes)

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        schema_element_type = self.stack.top.schema_element_type

        definition = None
        recurse = True
        if schema_element_type == "properties":
            parent = self.stack[-2].data
            definition = parent.get(OAREPO_MARSHMALLOW_PROPERTY, {})

            # set nested if not already set
            if "nested" not in definition:
                definition["nested"] = True

            generate_schema_class = definition.get("generate", True)
            schema_class = None  # to make pycharm happy
            schema_class_base_classes = None
            if generate_schema_class:
                if "class" not in definition:
                    definition["class"] = self.stack.top.key.title()
            if "class" in definition:
                schema_class = definition["class"]
                if "." not in schema_class:
                    schema_class = package_name(self.settings.python.record_schema_class) + "." + schema_class
                schema_class_base_classes = definition.get("base-classes", ["ma.Schema"])

            if generate_schema_class:
                self.marshmallow_stack.append(
                    MarshmallowNode(schema_class, schema_class_base_classes, self.stack.top.data)
                )
                # add nested to the definition
            else:
                schema_element_type = "nodef"
            recurse = generate_schema_class

        if recurse:
            # process children
            self.build_children()

        data = self.stack.top.data

        if not self.marshmallow_stack:
            return

        marshmallow_stack_top = self.marshmallow_stack[-1]

        if schema_element_type == "nodef":
            definition = self.get_marshmallow_definition(data, self.stack, definition=definition)
            self.marshmallow_stack[-1].prepared_schema = definition
        elif schema_element_type == "items":
            definition = self.get_marshmallow_definition(data, self.stack)
            definition["field"] = f"ma_fields.List({definition['field']})"
            self.marshmallow_stack[-1].prepared_schema = definition
        elif schema_element_type == "property":
            prepared_schema = marshmallow_stack_top.pop_prepared_schema()
            if prepared_schema:
                deepmerge(data.setdefault(OAREPO_MARSHMALLOW_PROPERTY, {}), prepared_schema)
            definition = self.get_marshmallow_definition(data, self.stack)
            key = definition.get("field_name", self.stack.top.key)
            marshmallow_stack_top.add(key, definition["field"])
        elif schema_element_type == "properties":
            node = self.marshmallow_stack.pop()
            class_name = node.schema_class_name
            if class_name.startswith("."):
                class_name = resolve_relative_classname(class_name, self.settings.python[self.class_config])
            if class_name not in self.generated_classes:
                self.generated_classes.add(class_name)
                self.process_template(
                    self.class_to_path(class_name),
                    "object-schema",
                    schema_class=class_name,
                    schema_bases=node.schema_class_bases,
                    fields=node.fields,
                    imports=self.imports,
                    imported_classes=self.imported_classes,
                    current_package_name=package_name(class_name),
                )

    def get_marshmallow_definition(self, data, stack, definition=None):
        if not definition:
            definition = data.setdefault(OAREPO_MARSHMALLOW_PROPERTY, {})

        definition = self.call_components("invenio_before_set_marshmallow_definition", definition, stack=stack)

        data_type = data.get("type", "object")
        generator = self.schema.settings.python.marshmallow.mapping.get(data_type, None)
        if isinstance(generator, dict):
            # a value, not a callable => merge it
            definition = deepmerge(generator, definition)
            data["marshmallow"] = definition
            generator = None

        # add imports if required
        for imp in definition.get("imports", []):
            self.imports[imp["import"]].add(imp["alias"])

        self.imported_classes.update(definition.get("imported-classes", {}))

        if "field" in definition:
            # do not modify the field
            return definition

        if "class" in definition:
            class_name = definition["class"]
            if class_name.startswith("."):
                class_name = resolve_relative_classname(class_name, self.settings.python[self.class_config])
            if "." in class_name:
                if not with_defined_prefix(self.settings.python.always_defined_import_prefixes, class_name):
                    class_base_name = self.imported_classes[class_name] = base_name(class_name)
                else:
                    class_base_name = class_name
            else:
                class_base_name = class_name

            # generate instance of the class, filling the options and validators

            definition["field"] = create_field(class_base_name, options=(), validators=(), definition=definition)
            return definition

        # if no generator from settings, get the default one
        if not generator:
            generator = default_marshmallow_generators.get(data_type, None)
            if not generator:
                if data_type == "object":
                    raise Exception(
                        f'Do not have marshmallow field generator for type "{data_type}" at path "{stack.path}". '
                        f"Either supply an existing schema class or instruct the compiler to generate one. "
                        f"See the documentation (docs/model.md, section oarepo:marshmallow) for details."
                    )
                raise Exception(
                    f'Do not have marshmallow field generator for type "{data_type}" at path "{stack.path}". '
                    f"Define it either in invenio_record_schema.py or in your own config"
                )

        # and generate the field
        definition["field"] = generator(data, definition, self.schema, self.imports)
        definition = self.call_components("invenio_after_set_marshmallow_definition", definition, stack=stack)
        return definition


def create_field(field_type, options=(), validators=(), definition=None):
    opts = [*options, *definition.get("options", [])]
    validators = [*validators, *definition.get("validators", [])]
    nested = definition.get("nested", False)
    list_nested = definition.get("list_nested", False)
    if validators:
        opts.append(f'validate=[{",".join(validators)}]')
    kwargs = definition.get("field_args", "")
    if kwargs and opts:
        kwargs = ", " + kwargs
    if not nested:
        ret = f'{field_type}({", ".join(opts)}{kwargs})'
    else:
        ret = f"{field_type}()"
    if nested:
        if opts or kwargs:
            ret = f'ma_fields.Nested(lambda: {ret}, {", ".join(opts)}{kwargs})'
        else:
            ret = f"ma_fields.Nested(lambda: {ret})"
    if list_nested:
        if opts or kwargs:
            ret = f'ma_fields.List(ma_fields.Nested(lambda: {ret}, {", ".join(opts)}{kwargs}))'
        else:
            ret = f"ma_fields.List(ma_fields.Nested(lambda: {ret}))"
    return ret


def marshmallow_string_generator(data, definition, schema, imports):
    validators = []
    min_length = data.get("minLength", None)
    max_length = data.get("maxLength", None)
    if min_length is not None or max_length is not None:
        validators.append(f"ma_valid.Length(min={min_length}, max={max_length})")
    return create_field("ma_fields.String", [], validators, definition)


def marshmallow_integer_generator(data, definition, schema, imports):
    return marshmallow_generic_number_generator("ma_fields.Integer", data, definition, schema, imports)


def marshmallow_number_generator(data, definition, schema, imports):
    return marshmallow_generic_number_generator("ma_fields.Float", data, definition, schema, imports)


def marshmallow_generic_number_generator(datatype, data, definition, schema, imports):
    validators = []
    minimum = data.get("minimum", None)
    maximum = data.get("maximum", None)
    exclusive_minimum = data.get("exclusiveMinimum", None)
    exclusive_maximum = data.get("exclusiveMaximum", None)

    if minimum is not None or maximum is not None or exclusive_minimum is not None or exclusive_maximum is not None:
        validators.append(
            f"ma_valid.Range("
            f'min={minimum or exclusive_minimum or "None"}, max={maximum or exclusive_maximum or "None"}, '
            f"min_inclusive={exclusive_minimum is None}, max_inclusive={exclusive_maximum is None})"
        )
    return create_field(datatype, [], validators, definition)


# TODO: rest of supported schema types

default_marshmallow_generators = {
    "string": marshmallow_string_generator,
    "integer": marshmallow_integer_generator,
    "number": marshmallow_number_generator,
}


def resolve_relative_classname(class_name, base_class_name):
    base_class_name = base_class_name.rsplit(".", maxsplit=1)[0]
    class_name = class_name[1:]
    if "." in class_name:
        # always go one level up
        class_name = "." + class_name
    while class_name.startswith("."):
        class_name = class_name[1:]
        base_class_name = base_class_name.rsplit(".", maxsplit=1)[0]
    return base_class_name + "." + class_name



from pathlib import Path

import pkg_resources


class TemplateRegistry:
    def __init__(self):
        self.mapping = {}
        for ep in reversed(
            sorted(
                pkg_resources.iter_entry_points("oarepo_model_builder.templates"),
                key=lambda ep: ep.name,
            )
        ):
            loaded_package = ep.load()
            base_path = Path(loaded_package.__file__).parent.absolute()
            for k, v in loaded_package.TEMPLATES.items():
                self.mapping[k] = base_path.joinpath(v)

    def get_template(self, template_key, settings):
        # try to get the template key from settings
        path = settings.python.templates.get(template_key, self.mapping.get(template_key, None))
        if not path:
            raise AttributeError(f"Template with key {template_key} has not been found")
        if isinstance(path, str):
            path = Path(path).absolute()
        if path.exists():
            with path.open() as f:
                return f.read()
        raise AttributeError(f"Template with key {template_key} has not been found, file at path {path} does not exist")


templates = TemplateRegistry()



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioRecordSearchPoetryBuilder(OutputBuilder):
    TYPE = "invenio_record_search_poetry"

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")

        output.set(
            "tool.poetry.plugins.'invenio_search.mappings'",
            self.settings.python.record_mapping_poetry,
            self.settings.mapping_package,
        )



from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.camelcase import camel_case


class MarshmallowClassGeneratorPreprocessor(PropertyPreprocessor):
    TYPE = "marshmallow_class_generator"

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: stack.schema_valid,
    )
    def modify_object_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        schema_element_type = stack.top.schema_element_type

        if schema_element_type == "properties":
            for k, v in stack.top.data.items():
                self.add_class_name(stack, k, v)
        elif schema_element_type == "items":
            self.add_class_name(stack, None, stack.top.data)
        return data

    def add_class_name(self, stack, key, data):
        if "items" not in data and "properties" not in data:
            return
        definition = data.setdefault("oarepo:marshmallow", {})
        if "class" in definition:
            return
        definition["class"] = camel_case(key or self.get_property_name(stack)) + "Schema"

    def get_property_name(self, stack):
        for el in reversed(stack.stack):
            if el.schema_element_type == "property":
                return el.key
        return "UnknownName"



import datetime
import logging
import os
import sys
import traceback
from pathlib import Path

import click

from oarepo_model_builder.conflict_resolvers import AutomaticResolver, InteractiveResolver
from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.utils.verbose import log


@click.command()
@click.option(
    "--output-directory",
    help="Output directory where the generated files will be placed. " 'Defaults to "."',
)
@click.option(
    "--package",
    help="Package into which the model is generated. "
    "If not passed, the name of the current directory, "
    "converted into python package name, is used.",
)
@click.option(
    "--set",
    "sets",
    help="Overwrite option in the model file. Example " "--set name=value",
    multiple=True,
)
@click.option(
    "-v",
    "verbosity",
    help="Increase the verbosity. This option can be used multiple times.",
    count=True,
)
@click.option(
    "--config",
    "configs",
    help="Load a config file and replace parts of the model with it. "
    "The config file can be a json, yaml or a python file. "
    "If it is a python file, it is evaluated with the current "
    'model stored in the "oarepo_model" global variable and '
    "after the evaluation all globals are set on the model.",
    multiple=True,
)
@click.option("--isort/--skip-isort", default=True, help="Call isort on generated sources")
@click.option("--black/--skip-black", default=True, help="Call black on generated sources")
@click.option("--resolve-conflicts", type=click.Choice(["replace", "keep", "comment", "debug"]))
@click.argument("model_filename")
def run(output_directory, package, sets, configs, model_filename, verbosity, isort, black, resolve_conflicts):
    """
    Compiles an oarepo model file given in MODEL_FILENAME into an Invenio repository model.
    """
    try:
        run_internal(
            output_directory, model_filename, package, configs, resolve_conflicts, sets, black, isort, verbosity
        )
    except Exception as e:
        if verbosity >= 2:
            print(e, file=sys.stderr)
            traceback.print_exc()
        elif verbosity >= 1:
            print(e, file=sys.stderr)
            traceback.print_exc(limit=-5)
        else:
            print(e, file=sys.stderr)
        sys.exit(1)


def run_internal(output_directory, model_filename, package, configs, resolve_conflicts, sets, black, isort, verbosity):
    # extend system's search path to add script's path in front (so that scripts called from the compiler are taken
    # from the correct virtual environ)
    os.environ["PATH"] = str(Path(sys.argv[0]).parent.absolute()) + os.pathsep + os.environ.get("PATH", "")
    if not output_directory:
        output_directory = os.getcwd()
    # set the logging level, it will be warning - 1 (that is, 29) if not verbose,
    # so that warnings only will be emitted. With each verbosity level
    # it will decrease
    logging.basicConfig(level=logging.INFO - verbosity, format="")
    Path(output_directory).mkdir(parents=True, exist_ok=True)
    handler = logging.FileHandler(Path(output_directory) / "installation.log", "a")
    handler.setLevel(logging.INFO)
    logging.root.addHandler(handler)
    log.enter(
        0,
        "\n\n%s\n\nProcessing model %s into output directory %s",
        datetime.datetime.now(),
        model_filename,
        output_directory,
    )
    model = load_model(model_filename, package, configs, black, isort, sets)
    model.schema["output-directory"] = output_directory
    if not resolve_conflicts or resolve_conflicts == "debug":
        resolver = InteractiveResolver(resolve_conflicts == "debug")
    else:
        resolver = AutomaticResolver(resolve_conflicts)
    builder = create_builder_from_entrypoints()
    builder.build(model, output_directory, resolver)
    log.leave("Done")
    print(f"Log saved to {Path(output_directory) / 'installation.log'}")


if __name__ == "__main__":
    run()



import sys
from pathlib import Path

import pkg_resources

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch


def create_builder_from_entrypoints(**kwargs):
    output_classes = load_entry_points_list("oarepo_model_builder.outputs")
    builder_classes = load_entry_points_list("oarepo_model_builder.builders")
    preprocess_classes = load_entry_points_list("oarepo_model_builder.property_preprocessors")
    model_preprocessor_classes = load_entry_points_list("oarepo_model_builder.model_preprocessors")

    builder_types = [x.TYPE for x in builder_classes]
    output_builder_components = {
        builder_type: load_entry_points_list(f"oarepo_model_builder.builder_components.{builder_type}")
        for builder_type in builder_types
    }

    return ModelBuilder(
        output_builders=builder_classes,
        outputs=output_classes,
        property_preprocessors=preprocess_classes,
        model_preprocessors=model_preprocessor_classes,
        output_builder_components=output_builder_components,
        **kwargs,
    )


def load_entry_points_dict(name):
    return {ep.name: ep.load() for ep in pkg_resources.iter_entry_points(group=name)}


def load_entry_points_list(name):
    ret = [(ep.name, ep.load()) for ep in pkg_resources.iter_entry_points(group=name)]
    ret.sort()
    return [x[1] for x in ret]


def load_model_from_entrypoint(ep: pkg_resources.EntryPoint):
    def load(schema):
        filename = ".".join(ep.attrs)
        data = pkg_resources.resource_string(ep.module_name, filename)
        return schema._load(filename, content=data)

    return load


def load_included_models_from_entry_points():
    ret = {}
    for ep in pkg_resources.iter_entry_points(group="oarepo.models"):
        ret[ep.name] = load_model_from_entrypoint(ep)
    return ret


def load_model(
    model_filename,
    package=None,
    configs=(),
    black=True,
    isort=True,
    sets=(),
    model_content=None,
    extra_included=None,
):
    loaders = load_entry_points_dict("oarepo_model_builder.loaders")
    included_models = load_included_models_from_entry_points()
    if extra_included:
        included_models.update(extra_included)
    schema = ModelSchema(
        model_filename,
        content=model_content,
        loaders=loaders,
        included_models=included_models,
    )
    for config in configs:
        load_config(schema, config, loaders)
    for s in sets:
        k, v = s.split("=", 1)
        schema.schema[k] = v
    check_plugin_packages(schema)
    if package:
        schema.settings["package"] = package
    if "python" not in schema.settings:
        schema.settings.python = HyphenMunch()
    schema.settings.python.use_isort = isort
    schema.settings.python.use_black = black
    return schema


def load_config(schema, config, loaders):
    old_loaders = schema.loaders
    schema.loaders = loaders
    try:
        loaded_file = schema._load(config)
        schema.merge(loaded_file)
    finally:
        schema.loaders = old_loaders


def check_plugin_packages(schema):
    try:
        required_packages = schema.schema.plugins.packages
    except AttributeError:
        return
    import subprocess

    import pkg_resources

    known_packages = set(d.project_name for d in pkg_resources.working_set)
    unknown_packages = [rp for rp in required_packages if rp not in known_packages]
    if unknown_packages:
        if (
            input(
                f'Required packages {", ".join(unknown_packages)} are missing. '
                f"Should I install them for you via pip install? (y/n) "
            )
            == "y"
        ):
            if subprocess.call(["pip", "install", *unknown_packages]):
                sys.exit(1)
            print("Installed required packages, please run this command again")
        sys.exit(1)



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput
from ..utils.verbose import log


class InvenioSampleAppPoetryBuilder(OutputBuilder):
    TYPE = "invenio_sample_app_poetry"

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")

        output.setdefault(
            "tool.poetry.dependencies.invenio",
            "version",
            "^3.5.0a3",
            "extras",
            ["base", "auth", "metadata", "files", "postgresql", "elasticsearch7"],
            "optional",
            True,
            "allow-prereleases",
            True,
        )

        output.setdefault("tool.poetry.dependencies.pyyaml", "version", ">=6", "optional", True)

        output.setdefault(
            "tool.poetry.dependencies.invenio-rest",
            "version",
            "^1.2.8",
            "optional",
            True,
            "allow-prereleases",
            True,
        )

        output.setdefault(
            "tool.poetry.dependencies.invenio-records-resources",
            "version",
            "^0.18.3",
            "optional",
            True,
            "allow-prereleases",
            True,
        )

        output.setdefault(
            "tool.poetry.extras",
            "sample-app",
            ["invenio", "invenio-records-resources", "pyyaml"],
        )

        if output.created:
            log(
                log.INFO,
                f"""To install the sample app, run
    poetry install -E sample-app            
            """,
            )



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class DatePreprocessor(PropertyPreprocessor):
    TYPE = "date"

    @process(
        model_builder=JSONSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "date",
    )
    def modify_date_jsonschema(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        data.setdefault("format", "date")
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "date",
    )
    def modify_date_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        deepmerge(
            data.setdefault("oarepo:marshmallow", {}),
            {
                "class": "ma_fields.Date"
                # TODO: validators, such as minDate, maxDate, minDateExclusive, maxDateExclusive
            },
        )
        return data



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioExtPoetryBuilder(OutputBuilder):
    TYPE = "invenio_ext_poetry"

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")

        ext_class = self.settings.python.ext_class.rsplit(".", maxsplit=1)

        output.set(
            "tool.poetry.plugins.'invenio_base.api_apps'",
            self.settings.package,
            f"{ext_class[0]}:{ext_class[-1]}",
        )



import lazy_object_proxy
from libcst import (
    Arg,
    Assign,
    Call,
    ClassDef,
    Dict,
    Element,
    Expr,
    FunctionDef,
    Import,
    ImportFrom,
    Integer,
    List,
    Module,
    Name,
    Pass,
    SimpleStatementLine,
    SimpleString,
    Tuple,
)


@lazy_object_proxy.Proxy
def module_mergers():
    from .indented_nodes import ModuleMerger

    return {Module: ModuleMerger()}


@lazy_object_proxy.Proxy
def indented_block_mergers():
    from .indented_nodes import ClassMerger
    from .simple_nodes import FunctionMerger, SimpleStatementLineMerger

    return {
        SimpleStatementLine: SimpleStatementLineMerger(),
        ClassDef: ClassMerger(),
        FunctionDef: FunctionMerger(),
    }


@lazy_object_proxy.Proxy
def simple_line_mergers():
    from .simple_nodes import AssignMerger, ExprMerger, ImportFromMerger, ImportMerger, PassMerger

    return {
        Assign: AssignMerger(),
        Import: ImportMerger(),
        ImportFrom: ImportFromMerger(),
        Expr: ExprMerger(),
        Pass: PassMerger(),
    }


@lazy_object_proxy.Proxy
def call_mergers():
    from .call import ArgMerger

    return {Arg: ArgMerger()}


@lazy_object_proxy.Proxy
def expression_mergers():
    from oarepo_model_builder.utils.cst.collections import DictMerger, ElementMerger, ListMerger

    from .call import CallMerger
    from .simple_nodes import ExprMerger, IntegerMerger, NameMerger, SimpleStringMerger

    return {
        Call: CallMerger(),
        List: ListMerger(),
        Tuple: ListMerger(),
        Integer: IntegerMerger(),
        Element: ElementMerger(),
        SimpleString: SimpleStringMerger(),
        Name: NameMerger(),
        Expr: ExprMerger(),
        Dict: DictMerger(),
    }



from typing import Callable, List

import faker
import faker.providers
from jinja2 import Environment, FunctionLoader

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.builders.utils import ensure_directory
from oarepo_model_builder.templates import templates

from ..builder import ModelBuilder
from ..builders import process
from ..builders.json_base import JSONBaseBuilder
from ..entrypoints import load_entry_points_list
from ..property_preprocessors import PropertyPreprocessor


class SampleDataGenerator(faker.Generator):
    def __init__(self, **config):
        super().__init__(**config)
        self.formatters = set()

    def set_formatter(self, name: str, method: Callable) -> None:
        self.formatters.add(name)
        return super().set_formatter(name, method)


# provider for Faker
class Provider:
    def __init__(self, generator) -> None:
        self.generator = generator

    def random_float(self):
        rnd = self.generator.random.randrange(-100, 100 + 1, 1)
        return rnd / 10

    def sample_object(self):
        return {self.generator.word(): self.generator.word() for _ in range(self.generator.random.randrange(1, 5, 1))}

    def constant(self, value=None):
        return value


SKIP = "skip"


class InvenioScriptSampleDataBuilder(JSONBaseBuilder):
    TYPE = "script_sample_data"
    output_file_type = "yaml"
    output_file_name = "script-import-sample-data"
    parent_module_root_name = "jsonschemas"

    def __init__(self, builder: ModelBuilder, property_preprocessors: List[PropertyPreprocessor]):
        super().__init__(builder, property_preprocessors)
        self.generator = SampleDataGenerator()
        from faker.config import PROVIDERS

        self.faker = faker.Faker(
            generator=self.generator,
            providers=[
                "oarepo_model_builder.invenio.invenio_script_sample_data",
                *PROVIDERS,
            ],
        )

        self.sample_data_providers = load_entry_points_list("oarepo_model_builder.sample_data_providers") + [
            faker_provider
        ]

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def model_element(self):
        schema_element_type = self.stack.top.schema_element_type

        if schema_element_type == "property":
            self.generate_property(self.stack.top.key)
        elif schema_element_type == "items":
            # the count is in the oarepo:sample section above the "items" element, so need to look at [-2], not the top
            count = self.get_count(self.stack[-2].data, None)
            if count is None:
                count = self.faker.random_int(1, 5)
            for key in range(count):
                self.generate_property(key)
        else:
            self.build_children()

    def generate_property(self, key):
        if not self.skip(self.stack):
            if "properties" in self.stack.top.data:
                self.output.enter(key, {})
                self.build_children()
                self.output.leave()
            elif "items" in self.stack.top.data:
                self.output.enter(key, [])
                self.build_children()
                self.output.leave()
            else:
                self.output.primitive(key, self.generate_fake(self.stack))

    def build(self, schema):
        output_name = schema.settings[self.output_file_name]
        output = self.builder.get_output(self.output_file_type, output_name)
        if not output.created:
            return
        count = self.get_count(schema.schema)
        for _ in range(count):
            super().build(schema)

    def get_count(self, schema, default_count=50):
        return schema.get("oarepo:sample", {}).get("count", default_count)

    def skip(self, stack):
        return get_oarepo_sample(stack).get("skip", False)

    def on_enter_model(self, output_name):
        self.output.next_document()

    def generate_fake(self, stack):
        params = {}
        method = None
        config = get_oarepo_sample(stack)
        params = config.get("params", params)

        for provider in self.sample_data_providers:
            ret = provider(self.faker, self.settings, stack, params)
            if ret is not SKIP:
                return ret


def get_oarepo_sample(stack):
    if isinstance(stack.top.data, dict) and "oarepo:sample" in stack.top.data:
        sample = stack.top.data.get("oarepo:sample")
        if isinstance(sample, dict):
            return sample
        if isinstance(sample, str):
            return {"faker": "constant", "params": {"value": sample}}
        if isinstance(sample, (list, dict)):
            if stack.top.schema_element_type == "property" and "items" in stack.top.data:
                return {"faker": "random_elements", "params": {"elements": sample, "unique": True}}
            else:
                return {"faker": "random_element", "params": {"elements": sample}}

    return {}


def faker_provider(faker, settings, stack, params):
    config = get_oarepo_sample(stack)
    method = config.get("faker")
    if not method:
        if stack.top.key in faker.formatters:
            method = stack.top.key
        else:
            data_type = stack.top.data.get("type")
            if data_type == "integer":
                method = "random_int"
            elif data_type == "number":
                method = "random_float"
            elif data_type == "date":
                method = "date"
            elif data_type == "object":
                # it is an object with unknown properties, return a sample object
                method = "sample_object"
            elif data_type in ("string", "keyword", "fulltext", "fulltext+keyword"):
                method = "sentence"
            else:
                print(
                    f"Warning: do not know how to generate sample data for {data_type} at path {stack.path}, using plain string rule"
                )
                method = "sentence"
    return getattr(faker, method)(**params)


class InvenioScriptSampleDataShellBuilder(OutputBuilder):
    TYPE = "invenio_script_runserver"

    def finish(self):
        context = {"settings": self.schema.settings}

        env = Environment(
            loader=FunctionLoader(lambda tn: templates.get_template(tn, context["settings"])),
            autoescape=False,
        )

        ensure_directory(self.builder, "scripts")
        output = self.builder.get_output("diff", "scripts/load_sample_data.sh")
        output.write(env.get_template("script-import-sample-data-shell").render(context))
        output.make_executable()



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput
from ..utils.verbose import log


class PoetryBuilder(OutputBuilder):
    TYPE = "poetry"

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")

        output.setdefault(
            "tool.poetry",
            "name",
            self.settings.package_base.replace("_", "-"),
            "version",
            "0.0.1",
            "description",
            f"A sample application for {self.settings.package}",
            "authors",
            [],
        )

        output.setdefault(
            "build-system",
            "requires",
            ["poetry-core>=1.0.0"],
            "build-backend",
            "poetry.core.masonry.api",
        )

        output.setdefault("tool.poetry.dependencies", "python", "^3.9")

        if "runtime-dependencies" in self.schema:
            output.table("tool.poetry.dependencies")
            for dep, value in self.schema.runtime_dependencies.items():
                if isinstance(value, str):
                    value = {"version": value}
                output.setdefault(
                    "tool.poetry.dependencies." + dep,
                    *sum((list(x) for x in value.items()), []),
                )

        if "dev-dependencies" in self.schema:
            output.table("tool.poetry.dev-dependencies")
            for dep, value in self.schema.dev_dependencies.items():
                if isinstance(value, str):
                    value = {"version": value}
                output.setdefault(
                    "tool.poetry.dev-dependencies." + dep,
                    *sum((list(x) for x in value.items()), []),
                )

        if output.created:
            log(
                log.INFO,
                f"""To install the sample app, run
    poetry install -E sample-app            
            """,
            )



from oarepo_model_builder.utils.cst import PythonContext
from oarepo_model_builder.utils.cst.common import IdentityMerger, MergerBase
from oarepo_model_builder.utils.cst.mergers import expression_mergers


class ElementMerger(MergerBase):
    """list element"""

    def identity(self, context: PythonContext, node):
        node_value = node.value
        merger = expression_mergers.get(type(node_value), IdentityMerger())
        return merger.identity(context, node_value)

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        existing_value = existing_node.value
        new_value = new_node.value

        merger = expression_mergers.get(type(existing_value), IdentityMerger())
        return existing_node.with_changes(value=merger.merge(context, existing_value, new_value))


class ListMerger(MergerBase):
    def identity(self, context: PythonContext, node):
        return node

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return existing_node.with_changes(
            elements=merge_lists_remove_duplicates(
                context, existing_node.elements, new_node.elements, expression_mergers
            )
        )


class DictMerger(MergerBase):
    def identity(self, context: PythonContext, node):
        return node

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        ret = []
        mergers = expression_mergers
        existing_elements = {el.key.value: el for el in existing_node.elements}
        new_elements = {el.key.value: el for el in new_node.elements}
        for k, el in existing_elements.items():
            merger = mergers.get(type(el), IdentityMerger())
            if k not in new_elements:
                ret.append(merger.merge(context, el, None))
            else:
                new_element = new_elements.pop(k)
                ret.append(merger.merge(context, el, new_element))

        for k, el in new_elements.items():
            merger = mergers.get(type(el), IdentityMerger())
            merged = merger.merge(context, None, el)
            ret.append(merged)
        return existing_node.with_changes(elements=[x for x in ret if x is not context.REMOVED])


def merge_lists_remove_duplicates(context: PythonContext, existing_list, new_list, mergers):
    ret = []
    new_list = [*new_list]

    for e in existing_list:
        merger = mergers.get(type(e), IdentityMerger())
        e_identity = merger.identity(context, e)

        for idx, n in enumerate(new_list):
            if not isinstance(n, type(e)):
                continue
            n_identity = merger.identity(context, n)
            if e_identity.deep_equals(n_identity):
                ret.append(merger.merge(context, e, n))
                del new_list[idx]
                break
        else:
            ret.append(merger.merge(context, e, None))

    for n in new_list:
        merger = mergers.get(type(n), IdentityMerger())
        ret.append(merger.merge(context, None, n))

    return [x for x in ret if x is not context.REMOVED]



from libcst import Assign, ClassDef, Import, ImportFrom, SimpleStatementLine

from .common import IdentityMerger, MergerBase, PythonContext, node_with_type
from .mergers import indented_block_mergers


class PriorityMergerMixin:
    def extract_body(self, node):
        raise NotImplementedError()

    def get_node_type(self, node_type_category, node):
        if isinstance(node, SimpleStatementLine):
            node = node.body[0]
        t = type(node)
        return node_type_category.get(t, "unknown")

    def _merge_children_with_priorities(self, context, existing_node, updated_node, mergers, node_type_category):
        existing_list = self.extract_body(existing_node)
        new_list = self.extract_body(updated_node)

        ret = []
        existing_list = [node_with_type(e, self.get_node_type(node_type_category, e)) for e in existing_list]
        new_list = [node_with_type(e, self.get_node_type(node_type_category, e)) for e in new_list]

        last_type = None
        for existing in existing_list:
            merger = mergers.get(type(existing.node), IdentityMerger())
            if last_type is not None and last_type != existing.type:
                while new_list and new_list[0].type == last_type:
                    new_node = new_list.pop(0).node
                    new_merger = mergers.get(type(new_node), IdentityMerger())
                    ret.append(new_merger.merge(context, None, new_node))
            last_type = existing.type
            found = False
            for idx, new in enumerate(new_list):
                if new.type != existing.type:
                    break
                if isinstance(existing.node, type(new.node)):
                    if merger.identity(context, existing.node).deep_equals(merger.identity(context, new.node)):
                        ret.append(merger.merge(context, existing.node, new.node))
                        del new_list[idx]
                        found = True
                        break
            if not found:
                ret.append(merger.merge(context, existing.node, None))

        for node in new_list:
            merger = mergers.get(type(node.node), IdentityMerger())
            ret.append(merger.merge(context, None, node.node))

        return self.finalize(existing_node, ret)

    def finalize(self, existing_node, body):
        raise NotImplementedError()


class ModuleMerger(PriorityMergerMixin, MergerBase):
    def identity(self, context, node):
        return node

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return self._merge_children_with_priorities(
            context,
            existing_node,
            new_node,
            mergers=indented_block_mergers,
            node_type_category={
                Import: "import",
                ImportFrom: "import",
                ClassDef: "classdef",
                Assign: "assign",
            },
        )

    def extract_body(self, node):
        return node.body

    def finalize(self, existing_node, body):
        return existing_node.with_changes(body=body)


class ClassMerger(PriorityMergerMixin, MergerBase):
    def identity(self, context, node):
        return node.name

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return self._merge_children_with_priorities(
            context,
            existing_node,
            new_node,
            mergers=indented_block_mergers,
            node_type_category={
                Import: "import",
                ImportFrom: "import",
                ClassDef: "classdef",
                Assign: "assign",
            },
        )

    def extract_body(self, node):
        return node.body.body

    def finalize(self, node, new_content):
        return node.with_changes(body=node.body.with_changes(body=new_content))



from pathlib import Path

from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class JSONSchemaBuilder(JSONBaseBuilder):
    TYPE = "jsonschema"
    output_file_type = "jsonschema"
    output_file_name = "schema-file"
    parent_module_root_name = "jsonschemas"

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def model_element(self):
        self.model_element_enter()
        self.build_children()
        if self.stack.top.schema_element_type == "items" or self.stack.top.schema_element_type == "property":
            self.check_and_output_required()
        self.merge_jsonschema(self.stack.top.data)
        self.model_element_leave()

    def merge_jsonschema(self, data):
        if isinstance(data, dict) and "oarepo:jsonschema" in data:
            jsonschema = self.call_components("before_merge_jsonschema", data["oarepo:jsonschema"], stack=self.stack)
            self.output.merge_jsonschema(jsonschema)

    def check_and_output_required(self):
        top_data = self.stack.top.data
        if isinstance(top_data, dict) and "properties" in top_data:
            self.output.collect_required()

    def on_enter_model(self, output_name):
        ensure_parent_modules(self.builder, Path(output_name), ends_at=self.parent_module_root_name)

    def on_leave_model(self):
        self.check_and_output_required()



def package_name(value):
    return value.rsplit(".", maxsplit=1)[0]


def base_name(value):
    return value.rsplit(".", maxsplit=1)[-1]


def in_different_package(current_package_name, value):
    return current_package_name != package_name(value)


def with_defined_prefix(always_defined_import_prefixes, value):
    return package_name(value) in always_defined_import_prefixes



from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack


class TypeShortcutsPreprocessor(PropertyPreprocessor):
    TYPE = "type_shortcuts"

    @process(
        model_builder="*",
        path="/model/**",
        condition=lambda current, stack: stack.schema_valid,
    )
    def modify_type_shortcuts(self, data, stack: ModelBuilderStack, **kwargs):
        top = stack.top

        if not top.schema_element_type:
            return data

        self.set_type(top.schema_element_type, data)
        self.set_child_arrays(top.schema_element_type, data)

        return data

    def set_type(self, element_type, data):
        if not isinstance(data, dict):
            return

        if "type" in data:
            return

        if element_type in ("property", "items"):
            if "properties" in data or "additionalProperties" in data or "unprocessedProperties" in data:
                data["type"] = "object"
            elif "items" in data:
                data["type"] = "array"

    def set_child_arrays(self, element_type, data):
        if element_type == "properties":
            for k, v in list(data.items()):
                if k.endswith("[]"):
                    data.pop(k)
                    data[k[:-2]] = self.create_array(v)

    @staticmethod
    def create_array(value):
        ret = {}
        array = {}
        for k, v in value.items():
            if k.endswith("[]"):
                ret[k[:-2]] = v
            else:
                array[k] = v
        ret["type"] = "array"
        ret["items"] = array
        return ret

    @process(
        model_builder="*",
        path="/model",
        condition=lambda current, stack: stack.schema_valid,
    )
    def add_model_type(self, data, stack: ModelBuilderStack, **kwargs):
        if "type" not in data:
            data["type"] = "object"
        return data



from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch
from oarepo_model_builder.utils.jinja import package_name


class InvenioBaseClassPythonBuilder(PythonBuilder):
    class_config = None
    template = None
    parent_modules = True

    def finish(self, **extra_kwargs):
        python_path = self.class_to_path(self.settings.python[self.class_config])
        self.process_template(
            python_path,
            self.template,
            current_package_name=package_name(self.settings.python[self.class_config]),
            **extra_kwargs,
        )

    def process_template(self, python_path, template, **extra_kwargs):
        if self.parent_modules:
            self.create_parent_modules(python_path)
        output: PythonOutput = self.builder.get_output("python", python_path)
        context = HyphenMunch(settings=self.settings, python=self.settings.python, **extra_kwargs)
        template = self.call_components("invenio_before_python_template", template, context=context)
        output.merge(template, context)



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class TextKeywordPreprocessor(PropertyPreprocessor):
    TYPE = "text_keyword"
    #
    # type='fulltext' in model
    #

    @process(
        model_builder=JSONSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext",
    )
    def modify_fulltext_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data

    @process(
        model_builder=MappingBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext",
    )
    def modify_fulltext_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "text"
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext",
    )
    def modify_fulltext_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data

    #
    # type='keyword' in model
    #

    @process(
        model_builder=JSONSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "keyword",
    )
    def modify_keyword_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data

    @process(
        model_builder=MappingBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "keyword",
    )
    def modify_keyword_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "keyword"
        deepmerge(
            data.setdefault("oarepo:mapping", {}),
            {"ignore_above": self.settings["elasticsearch"]["keyword-ignore-above"]},
        )
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "keyword",
    )
    def modify_keyword_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data

    #
    # type='fulltext+keyword' in model
    #

    @process(
        model_builder=JSONSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext+keyword",
    )
    def modify_fulltext_keyword_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data

    @process(
        model_builder=MappingBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext+keyword",
    )
    def modify_fulltext_keyword_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "text"
        deepmerge(
            data.setdefault("oarepo:mapping", {}),
            {
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": self.settings["elasticsearch"]["keyword-ignore-above"],
                    }
                }
            },
            [],
        )
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "fulltext+keyword",
    )
    def modify_fulltext_keyword_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "string"
        return data



import tomlkit
import tomlkit.items
from tomlkit.exceptions import NonExistentKey

from oarepo_model_builder.outputs import OutputBase


class TOMLOutput(OutputBase):
    TYPE = "toml"

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = f.read()
                self.toml = tomlkit.parse(self.original_data)
                self.parsed = tomlkit.dumps(self.toml, sort_keys=True)
        except FileNotFoundError:
            self.original_data = None
            self.toml = tomlkit.document()
            self.parsed = None

    @property
    def created(self):
        return self.original_data is None

    def table(self, key, key_type=tomlkit.items.KeyType.Bare):
        def get_by_key(key):
            _key = create_key(key, key_type)
            try:
                return self.toml[_key]
            except KeyError:
                pass
            if "." not in key:
                return None
            if '"' in key:
                key = [*key.split('"')[:2]]
            elif "'" in key:
                key = [*key.split("'")[:2]]
            else:
                key = [key, None]

            key_seq = [x for x in key[0].split(".") if x]
            if key[1]:
                key_seq.append(key[1])

            t = self.toml
            for k in key_seq:
                _key = create_key(k, key_type)
                try:
                    t = t[_key]
                except KeyError:
                    return None
            return t

        t = get_by_key(key)
        if t is not None:
            return t
        t = tomlkit.table()
        self.toml.append(create_key(key, key_type), t)
        return t

    def get(self, table, key):
        try:
            tbl = self.toml[table]
            return tbl[key]
        except NonExistentKey:
            return None

    def set(self, table, key, value, *others_key_values, key_type=tomlkit.items.KeyType.Bare):
        tbl = self.table(table)
        key = create_key(key, key_type)
        tbl[key] = value
        while others_key_values:
            key = others_key_values[0]
            value = others_key_values[1]
            others_key_values = others_key_values[2:]

            key = create_key(key, key_type)
            tbl[key] = value

    def setdefault(self, table, key, value, *others_key_values, key_type=tomlkit.items.KeyType.Bare):
        tbl = self.table(table)

        key = create_key(key, key_type)
        tbl.setdefault(key, value)

        while others_key_values:
            key = others_key_values[0]
            value = others_key_values[1]
            others_key_values = others_key_values[2:]

            key = create_key(key, key_type)
            tbl.setdefault(key, value)

    def finish(self):
        out = tomlkit.dumps(self.toml, sort_keys=True)
        if out != self.parsed:
            with self.builder.filesystem.open(self.path, "w") as f:
                f.write(tomlkit.dumps(self.toml))


def create_key(key, key_type=tomlkit.items.KeyType.Bare):
    return tomlkit.items.SingleKey(key, t=key_type)



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioRecordResourcePoetryBuilder(OutputBuilder):
    TYPE = "invenio_record_resource_poetry"

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")

        register_function = self.settings.python.create_blueprint_from_app.rsplit(".", maxsplit=1)

        output.set(
            "tool.poetry.plugins.'invenio_base.api_blueprints'",
            self.settings.python.record_resource_blueprint_name,
            f"{register_function[0]}:{register_function[-1]}",
        )



from typing import Mapping

import munch


class HyphenMunch(munch.AutoMunch):
    def __setitem__(self, key, value):
        if isinstance(value, Mapping) and not isinstance(value, (munch.AutoMunch, munch.Munch)):
            value = munch.munchify(value, HyphenMunch)
        return super().__setitem__(key, value)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except:
            try:
                key = key.replace("_", "-")
                return super().__getitem__(key)
            except:
                key = key.replace("-", "_")
                return super().__getitem__(key)



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class NumberPreprocessor(PropertyPreprocessor):
    TYPE = "number"

    @process(
        model_builder=MappingBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "number",
    )
    def modify_date_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "double"
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "number",
    )
    def modify_date_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        deepmerge(data.setdefault("oarepo:marshmallow", {}), {"class": "ma_fields.Float"})
        return data



from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.utils.deepmerge import deepmerge


class ElasticsearchModelPreprocessor(ModelPreprocessor):
    TYPE = "elasticsearch"

    def transform(self, schema, settings):
        deepmerge(
            settings,
            {
                "elasticsearch": {
                    "version": "v7",
                    "keyword-ignore-above": 50,
                    "templates": {"v7": {"mappings": {"properties": {}}}},
                }
            },
        )
